/**
 * Proto实例结构
 */
interface ProtoInstance {
	toJSON(): { [k: string]: any };
}

/**
 * Proto类结构
 */
interface ProtoClass {
	new (): ProtoInstance;
	create(properties?: any): any;
	encode(message: any, writer?: protobuf.Writer): protobuf.Writer;
	encodeDelimited(message: any, writer?: protobuf.Writer): protobuf.Writer;
	decode(reader: (protobuf.Reader | Uint8Array), length?: number): any;
	decodeDelimited(reader: (protobuf.Reader | Uint8Array)): any;
	verify(message: { [k: string]: any }): (string | null);
	fromObject(object: { [k: string]: any }): any;
	toObject(message: any, options?: protobuf.IConversionOptions): { [k: string]: any };
}

/**
 * Proto发送信息回调方法
 */
interface ProtoSendFunc<T extends ProtoInstance> {
	(arg: T): void;
}

class SocketMgr {
	private static _instance: SocketMgr;
	private debugUrl: string = '172.16.14.8';
	private debugPort: number = 9001;
	private releaseUrl: string = '172.16.14.8';//
	private releasePort: number = 9001;//
	static get instance(): SocketMgr {
		if (!this._instance) {
			this._instance = new SocketMgr();
		}
		return this._instance;
	}

	get url(): string {
		return DEBUG ? this.debugUrl : this.releaseUrl;
	}
	get port(): number {
		return DEBUG ? this.debugPort : this.releasePort;
	}


	// 以下方法提供给WebSocket+protobuf使用

	private socketClient: egret.WebSocket;
	private heartBeatTimer: egret.Timer;
	public heartBeatTimes: number = 0;
	private reconnectTimer: egret.Timer;
	private connected = false;
	private reconnecting = false;
	private reconnectTimes: number = 0;
	private connectedCBs: { callback: Function, thisObj: any }[] = [];
	private reconnectCBs: { callback: Function, thisObj: any }[] = [];
	private errorCBs: { callback: Function, thisObj: any }[] = [];
	private needReconnect = false;
	private hideLoading = true;

	public constructor() {
	}

	/**
 * 发送protobuf消息
 * @param {number} protoId 协议号
 * @param {ProtoClass} protoClass proto类
 * @param {ProtoSendFunc} func 对proto实例对象进行赋值处理的方法
 */
	public sendProtobufMsg<T extends ProtoInstance>(protoId: number, protoClass?: ProtoClass, func?: ProtoSendFunc<T>): void {
		console.log(`[Send To Server]protoId:${protoId}`);
		let msg = new egret.ByteArray();
		if (protoClass) {
			let proto = new protoClass();
			if (func) {
				func(<T>proto);
			}
			let writer = protoClass.encode(proto);
			msg.writeUnsignedShort(writer.len);
			msg.writeUnsignedShort(protoId);
			if (writer.len > 0) {
				msg._writeUint8Array(writer.finish());
			}
		} else {
			msg.writeUnsignedShort(0);
			msg.writeUnsignedShort(protoId);
		}
		this.socketClient.writeBytes(msg);
	}

    /**
     * 处理protobuf消息
     * @param {} msg proto字节数组
     * @param {ProtoClass} protoClass proto类，解析结果模板
     * @returns {} 返回对应proto类实例对象
     */
	public handleProtobufMsg<T>(msg: egret.ByteArray | Uint8Array, protoClass: ProtoClass): T {
		if (msg == null || msg.length < 1) {
			console.error('byte数组长度小于1！');
			return undefined;
		}
		if (msg instanceof egret.ByteArray) {
			return protoClass.decode(msg.bytes);
		} else {
			return protoClass.decode(msg);
		}
	}

	openSocket(): void {
		this.socketClient = new egret.WebSocket();
		this.socketClient.addEventListener(egret.ProgressEvent.SOCKET_DATA, this.onReceive, this);
		this.socketClient.addEventListener(egret.Event.CONNECT, this.onConnect, this);
		this.socketClient.addEventListener(egret.Event.CLOSE, this.onClose, this);
		this.socketClient.addEventListener(egret.IOErrorEvent.IO_ERROR, this.onError, this);
		this.socketClient.type = egret.WebSocket.TYPE_BINARY;
		this.socketClient.connect(this.url, this.port);
	}
	/**
 * 连接上服务器的回调
 */
	private onConnect(): void {
		console.log("已连上服务器！");
		this.connected = true;
		this.reconnectTimes = 0;
		// 心跳
		if (this.heartBeatTimer) {
			this.heartBeatTimer.stop();
		}
		this.heartBeatTimer = new egret.Timer(5000, 0);
		this.heartBeatTimer.addEventListener(egret.TimerEvent.TIMER, this.onHeartBeat, this);
		this.heartBeatTimer.start();

		this.connectedCBs.forEach((v) => {
			v.callback.call(v.thisObj);
		});

		setTimeout(() => {
			GameMgr.instance.dispatchEvent(new SocketEvent(SocketEvent.RECONNECT_MSG));
		}, 0)

		if (this.reconnecting) {
			this.reconnecting = false;
		}

		UserMgr.instance.sendUserinfo();
	}

	private onReceive(ev: egret.Event): void {
		let byteArr = new egret.ByteArray();
		this.socketClient.readBytes(byteArr);
		let socketEv = new SocketEvent(SocketEvent.PROTO_MSG);
		socketEv.length = byteArr.readUnsignedShort();
		socketEv.protoId = byteArr.readUnsignedShort();
		console.log(`[Receive From Server]protoId:${socketEv.protoId}`);
		if (socketEv.protoId == Protocol.heartBeat) {
			this.heartBeatTimes = 0;
		}
		if (socketEv.length > 0) {
			// 注：客户端不需要处理断包粘包问题
			socketEv.message = new egret.ByteArray();
			byteArr.readBytes(socketEv.message, 0, socketEv.length);//byteArr.readAvailable);
		}
		GameMgr.instance.dispatchEvent(socketEv);
	}

	private onClose(): void {
		console.log("socket closed!");

		if (this.connected) {
			this.connected = false;
			this.heartBeatTimer.stop();
			this.heartBeatTimer = null;
		}

		this.reconnect();
	}

	private reconnect(): void {
		this.reconnectTimes++;
		if (this.reconnectTimes >= 3) {
			console.log('网络状况较差，请检查网络');
			return;
		}
		console.log("Reconnect login!");
		this.connectedCBs = [];
		this.errorCBs = [];
		this.reconnecting = true;
		this.socketClient.removeEventListener(egret.ProgressEvent.SOCKET_DATA, this.onReceive, this);
		this.socketClient.removeEventListener(egret.Event.CONNECT, this.onConnect, this);
		this.socketClient.removeEventListener(egret.Event.CLOSE, this.onClose, this);
		this.socketClient.removeEventListener(egret.IOErrorEvent.IO_ERROR, this.onError, this);
		this.socketClient = null;

		this.openSocket();
	}

	/**
 * 连接发生错误
 */
	private onError(res: any) {
		// console.error("网络连接异常中断");

		// this.errorCBs.forEach((v) => {
		// 	v.callback.call(v.thisObj, res.errMsg);
		// });

		// this.onClose();
	}

	private onHeartBeat(): void {
		if (this.heartBeatTimes >= 1) {
			this.onClose();
			this.heartBeatTimes = 0;
		} else {
			this.sendProtobufMsg(Protocol.heartBeat);
			this.heartBeatTimes++;
		}
	}

	/**
     * 添加连上服务器监听
     */
	public addConnectListener(callback: Function, thisObj: any): void {
		this.connectedCBs.push({ callback, thisObj });
	}

    /**
     * 添加连接错误监听
     */
	public addErrorListener(callback: Function, thisObj: any): void {
		this.errorCBs.push({ callback, thisObj });
	}

}